#include <errno.h>
#include <fcntl.h>
#include <malloc.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#include "avilib.h"

#define MAX_AUDIO_TRACKS 20
#define MAX_VIDEO_TRACKS 20

struct audio_t {
  int bitrate;
  int channels;
  int rate;
};

int            xaudio[MAX_AUDIO_TRACKS];
int            xvideo;

char *basename;

int verbose = 0;

void usage(char *fname) {
  fprintf(stdout, "Usage: %s [-v|-o outname|-a track|-nv] -i inname\n",
          fname);
  fprintf(stdout, "       -i inname   Use 'inname' as the source.\n");
  fprintf(stdout, "       -o outname  Use 'outname' as the base for " \
          "destination file names.\n");
  fprintf(stdout, "                   '-v0', '-v1', '-a0', '-a1'... will be " \
          "appended to this name.\n");
  fprintf(stdout, "                   Default: use 'inname'.\n");
  fprintf(stdout, "       -a track    Extract specified audio track. Can be " \
          "used more than once.\n");
  fprintf(stdout, "                   Default: extract all tracks.\n");
  fprintf(stdout, "       -na         Don't extract any audio track.\n");
  fprintf(stdout, "       -nv         Don't extract any video track.\n");
  fprintf(stdout, "                   Default: extract all tracks.\n");
  fprintf(stdout, "       -v          Increase verbosity.\n");
  fprintf(stdout, "       -h          Show this help.\n");
}

void process_avi(avi_t *avi) {
  int fdout;
  char *newName;
  long i, frames;
  char *buffer;
  int nread, nwritten;
  
  buffer = malloc(1024 * 768 * 3);
  if (buffer == NULL) {
    fprintf(stderr, "(%s) Failed to malloc %d bytes.\n", __FILE__,
            1024 * 768 * 3);
    exit(1);
  }
  if (verbose > 0) {
    fprintf(stdout, "width=%d\nheight=%d\nframes=%ld\nfps=%f\ncodec=%s\n", 
            AVI_video_width(avi), AVI_video_height(avi),
            AVI_video_frames(avi), AVI_frame_rate(avi),
            AVI_video_compressor(avi));
  }
  if (xvideo) {
    newName = malloc(strlen(basename) + 6);
    if (newName == NULL) {
      fprintf(stderr, "(%s) Failed to malloc %d bytes.\n", __FILE__,
              strlen(basename) + 6);
      exit(1);
    }
    sprintf(newName, "%s-v", basename);
    fdout = open(newName, O_WRONLY | O_CREAT | O_TRUNC,
                 S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
    if (fdout < 0) {
      fprintf(stderr, "(%s) Failed to create \"%s\" (%d, %s).\n",
        __FILE__, newName, errno, strerror(errno));
      exit(1);
    }
    for (i = 0, frames = AVI_video_frames(avi); i < frames; i++) {
      int key;
      nread = AVI_read_frame(avi, buffer, &key);
      if (nread < 0) {
        fprintf(stderr, "(%s) AVI_read_frame error.\n", __FILE__);
        AVI_print_error("failure");
        exit(1);
      }
      nwritten = write(fdout, buffer, nread);
      if (verbose > 1)
        fprintf(stdout, "%d\n", nwritten);
    }
    close(fdout);
    if (verbose > 1)
      fprintf(stdout, "------\n");
  }
  for (i = 0; i < MAX_AUDIO_TRACKS; i++)
    if (xaudio[i] && (i < AVI_audio_tracks(avi))) {
      if (AVI_set_audio_track(avi, i)) {
        fprintf(stderr, "(%s) Failed to set audio track %d.\n", __FILE__, i);
        exit(1);
      }
      newName = malloc(strlen(basename) + 6);
      if (newName == NULL) {
        fprintf(stderr, "(%s) Failed to malloc %d bytes.\n", __FILE__,
                strlen(basename) + 6);
        exit(1);
      }
      sprintf(newName, "%s-a%d", basename, i);
      fdout = open(newName, O_WRONLY | O_CREAT | O_TRUNC,
                   S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
      if (fdout < 0) {
        fprintf(stderr, "(%s) Failed to create \"%s\" (%d, %s).\n",
          __FILE__, newName, errno, strerror(errno));
        exit(1);
      }
      do {
        nread = AVI_read_audio(avi, buffer, 1024 * 768 * 3);
        if (nread > 0) {
          nwritten = write(fdout, buffer, nread);
          if (verbose > 1)
            fprintf(stdout, "%d\n", nwritten);
        }
      } while (nread > 0);
      close(fdout);
      if (verbose > 1)
        fprintf(stdout, "------\n");
    }
}

int main(int argc, char *argv[]) {

  int i, aparam = 0, track;
  avi_t *avi = NULL;

  memset(xaudio, 1, MAX_AUDIO_TRACKS);
  xvideo = 1;
  for (i = 1; i < argc; i++) {
    if (!strcmp(argv[i], "-i")) {
      if ((i + 1) == argc) {
        fprintf(stderr, "(%s) -i needs a file name.\n", __FILE__);
        return 1;
      }

      avi = AVI_open_input_file(argv[i + 1], 1);
      if (avi == NULL) {
        fprintf(stderr, "(%s) Could not open \"%s\".\'", __FILE__,
                argv[i + 1]);
        return 1;
      }
      if (basename == NULL)
        basename = argv[i + 1];
      i++;

    } else if (!strcmp(argv[i], "-v"))
      verbose++;
    else if (!strcmp(argv[i], "-o")) {
      if ((i + 1) == argc) {
        fprintf(stderr, "(%s) -o needs a file name.\n", __FILE__);
        return 1;
      }
      basename = argv[i + 1];
      i++;
    } else if (!strcmp(argv[i], "-h")) {
      usage(argv[0]);
      return 0;
    } else if (!strcmp(argv[i], "-a")) {
      if ((i + 1) == argc) {
        fprintf(stderr, "(%s) -a needs a track number.\n", __FILE__);
        return 1;
      }
      if (!aparam) {
        memset(xaudio, 0, MAX_AUDIO_TRACKS);
        aparam = 1;
      }
      track = strtol(argv[i + 1], NULL, 10);
      if ((track < 0) || (track >= MAX_AUDIO_TRACKS)) {
        fprintf(stderr, "(%s) Audio track range is 0..%d.\n", __FILE__,
                MAX_AUDIO_TRACKS);
        return 1;
      }
      xaudio[track] = 1;
      i++;
    } else if (!strcmp(argv[i], "-na"))
      memset(xaudio, 0, MAX_AUDIO_TRACKS);
    else if (!strcmp(argv[i], "-nv"))
      xvideo = 0;
    else {
      fprintf(stderr, "(%s) Unknown parameter \"%s\".\n", __FILE__, argv[i]);
      usage(argv[0]);
    }
  }
  if (avi == NULL) {
    fprintf(stderr, "(%s) No source file specified.\n", __FILE__);
    return 1;
  }
  
  process_avi(avi);

  AVI_close(avi);  
  return 0;
}
